import os
import re
import shutil
import urllib.parse
from logging import Logger, getLogger
from pathlib import Path
from typing import Dict, List

import yaml

from cumulusci.core.tasks import BaseTask
from cumulusci.utils import elementtree_parse_file
from cumulusci.utils.xml import metadata_tree

__location__ = os.path.dirname(os.path.realpath(__file__))


def metadata_sort_key(name):
    sections = []
    for section in re.split("[.|-]", name):
        sections.append(metadata_sort_key_section(section))

    key = "_".join(sections)
    key = key.replace("_", "Z")

    return key


def metadata_sort_key_section(name):
    prefix = "5"
    key = name

    # Sort namespace prefixed names last
    base_name = name
    if base_name.endswith("__c"):
        base_name = base_name[:-3]
    if "__" in base_name:
        prefix = "8"

    key = prefix + name
    return key


def process_common_components(response_messages: List, components: Dict):
    """Compare compoents in the api responce object with list of components and return common common components"""
    if not response_messages or not components:
        return components
    for message in response_messages:

        message_list = message.firstChild.nextSibling.firstChild.nodeValue.split("'")
        if len(message_list) > 1:
            component_type = message_list[1]
            message_txt = message_list[2]
            if "is not available in this organization" in message_txt:
                del components[component_type]
            elif "is unknown" in message_txt:
                component_type = message_list[0].split(" ")
                components[component_type[0]].remove(message_list[1])
                if len(components[component_type]) == 0:
                    del components[component_type]
            else:
                component_name = message_list[3]
                if component_name in components[component_type]:
                    components[component_type].remove(component_name)
                    if len(components[component_type]) == 0:
                        del components[component_type]
    return components


class MetadataParserMissingError(Exception):
    pass


class PackageXmlGenerator(object):
    def __init__(
        self,
        directory,
        api_version,
        package_name=None,
        managed=None,
        delete=None,
        install_class=None,
        uninstall_class=None,
        types=None,
        logger=None,
    ):
        with open(
            __location__ + "/metadata_map.yml", "r", encoding="utf-8"
        ) as f_metadata_map:
            self.metadata_map = yaml.safe_load(f_metadata_map)
        self.directory = directory
        self.api_version = api_version
        self.package_name = package_name
        self.managed = managed
        self.delete = delete
        self.install_class = install_class
        self.uninstall_class = uninstall_class
        self.types = types or []
        self.logger = logger

    def __call__(self):
        if not self.types:
            self.parse_types()
        return self.render_xml()

    def parse_types(self):
        for item in sorted(os.listdir(self.directory)):
            if item == "package.xml":
                continue
            if not os.path.isdir(self.directory + "/" + item):
                continue
            if item.startswith("."):
                continue
            config = self.metadata_map.get(item)
            if not config:
                raise MetadataParserMissingError(
                    "No parser configuration found for subdirectory %s" % item
                )

            for parser_config in config:
                options = parser_config.get("options") or {}
                parser = globals()[parser_config["class"]](
                    parser_config["type"],  # Metadata Type
                    self.directory + "/" + item,  # Directory
                    parser_config.get("extension", ""),  # Extension
                    self.delete,  # Parse for deletion?
                    self.logger,  # Logger
                    **options,  # Extra kwargs
                )
                self.types.append(parser)

    def render_xml(self):
        lines = []

        # Print header
        lines.append('<?xml version="1.0" encoding="UTF-8"?>')
        lines.append('<Package xmlns="http://soap.sforce.com/2006/04/metadata">')
        if self.package_name:
            package_name_encoded = urllib.parse.quote(self.package_name, safe=" ")
            lines.append("    <fullName>{0}</fullName>".format(package_name_encoded))

        if self.managed and self.install_class:
            lines.append(
                "    <postInstallClass>{0}</postInstallClass>".format(
                    self.install_class
                )
            )

        if self.managed and self.uninstall_class:
            lines.append(
                "    <uninstallClass>{0}</uninstallClass>".format(self.uninstall_class)
            )

        # Print types sections
        self.types.sort(key=lambda x: x.metadata_type.upper())
        for parser in self.types:
            type_xml = parser()
            if type_xml:
                lines.extend(type_xml)

        # Print footer
        lines.append("    <version>{0}</version>".format(self.api_version))
        lines.append("</Package>")

        return "\n".join(lines)


class BaseMetadataParser(object):
    def __init__(self, metadata_type, directory, extension, delete, logger=None):
        self.metadata_type = metadata_type
        self.directory = directory
        self.extension = extension
        self.delete = delete
        self.members = []
        self.logger: Logger = logger or getLogger(__file__)

        if self.delete:
            self.delete_excludes = self.get_delete_excludes()

    def __call__(self):
        self.parse_items()
        return self.render_xml()

    def get_delete_excludes(self):
        filename = os.path.join(
            __location__, "..", "..", "files", "delete_excludes.txt"
        )
        excludes = []
        with open(filename, "r", encoding="utf-8") as f:
            for line in f:
                excludes.append(line.strip())
        return excludes

    def parse_items(self):
        # Loop through items
        for item in sorted(os.listdir(self.directory)):
            # on Macs this file is generated by the OS. Shouldn't be in the package.xml
            if item.startswith("."):
                continue

            # Ignore the CODEOWNERS file which is special to Github
            if item in ["CODEOWNERS", "OWNERS"]:
                continue

            if item.endswith("-meta.xml"):
                continue

            if self.extension and not item.endswith("." + self.extension):
                continue

            if self.check_delete_excludes(item):
                continue

            self.parse_item(item)

    def check_delete_excludes(self, item):
        if not self.delete:
            return False
        if item in self.delete_excludes:
            return True
        return False

    def parse_item(self, item):
        members = self._parse_item(item)
        if members:
            for member in members:
                # Translate filename namespace tokens into in-file namespace tokens
                member = member.replace("___NAMESPACE___", "%%%NAMESPACE%%%")
                self.members.append(member)

    def _parse_item(self, item):
        "Receives a file or directory name and returns a list of members"
        raise NotImplementedError("Subclasses should implement their parser here")

    def strip_extension(self, filename):
        return filename.rsplit(".", 1)[0]

    def render_xml(self):
        output = []
        if not self.members:
            return
        output.append("    <types>")
        self.members.sort(key=metadata_sort_key)
        for member in self.members:
            if isinstance(member, bytes):
                member = member.decode("utf-8")
            output.append("        <members>{0}</members>".format(member))
        output.append("        <name>{0}</name>".format(self.metadata_type))
        output.append("    </types>")
        return output

    def strip_folder(self, component_list):
        for item in sorted(os.listdir(self.directory)):
            if item.startswith("."):
                continue

            # Ignore the CODEOWNERS file which is special to Github
            if item in ["CODEOWNERS", "OWNERS"]:
                continue

            if item.endswith("-meta.xml"):
                continue

            if self.extension and not item.endswith("." + self.extension):
                continue

            self._strip_component(item, component_list)

    def _strip_component(self, item, component_list):
        raise NotImplementedError(
            "Subclasses should implement their stripping component logic"
        )


class MetadataFilenameParser(BaseMetadataParser):
    def _parse_item(self, item):
        return [self.strip_extension(item)]

    def _strip_component(self, item, component_list):
        if self.strip_extension(item) not in component_list:
            path = os.path.join(self.directory, item)
            self.logger.info(
                f"Deleting component {self.strip_extension(item)} of type {self.metadata_type}"
            )
            os.remove(path)
            if os.path.exists(path + "-meta.xml"):
                os.remove(path + "-meta.xml")


class MetadataFolderParser(BaseMetadataParser):
    def _parse_item(self, item):
        members = []
        path = self.directory + "/" + item

        # Skip non-directories
        if not os.path.isdir(path):
            return members

        # Only add the folder itself if its -meta.xml is present
        # (If there's no -meta.xml, this package is adding items to an existing folder.)
        if Path(path + "-meta.xml").exists():
            members.append(item)

        for subitem in sorted(os.listdir(path)):
            if subitem.endswith("-meta.xml") or subitem.startswith("."):
                continue
            submembers = self._parse_subitem(item, subitem)
            members.extend(submembers)

        return members

    def check_delete_excludes(self, item):
        return False

    def _parse_subitem(self, item, subitem):
        return [item + "/" + self.strip_extension(subitem)]

    def _strip_component(self, item, component_list):
        path = os.path.join(self.directory, item)
        # Check if it is a directory and take action
        if (
            os.path.isdir(path)
            and os.path.exists(path + "-meta.xml")
            and item not in component_list
        ):
            self.logger.info(f"Deleting component {item} of type {self.metadata_type}")
            shutil.rmtree(path)
            os.remove(path + "-meta.xml")

        if os.path.isdir(path):
            for subitem in sorted(os.listdir(path)):
                if subitem.endswith("-meta.xml") or subitem.startswith("."):
                    continue
                self._strip_subitem(item, subitem, component_list)

    def _strip_subitem(self, item, subitem, component_list):
        if item + "/" + self.strip_extension(subitem) not in component_list:
            path = os.path.join(self.directory, item, subitem)
            self.logger.info(
                f"Deleting component {item}/{self.strip_extension(subitem)} of type {self.metadata_type}"
            )
            os.remove(path)
            if os.path.exists(path + "-meta.xml"):
                os.remove(path + "-meta.xml")


class MissingNameElementError(Exception):
    pass


class ParserConfigurationError(Exception):
    pass


class MetadataXmlElementParser(BaseMetadataParser):

    namespaces = {"sf": "http://soap.sforce.com/2006/04/metadata"}

    def __init__(
        self,
        metadata_type,
        directory,
        extension,
        delete,
        logger=None,
        item_xpath=None,
        name_xpath=None,
    ):
        super(MetadataXmlElementParser, self).__init__(
            metadata_type, directory, extension, delete, logger
        )
        if not item_xpath:
            raise ParserConfigurationError("You must provide a value for item_xpath")
        self.item_xpath = item_xpath
        if not name_xpath:
            name_xpath = "./sf:fullName"
        self.name_xpath = name_xpath

    def _parse_item(self, item):
        root = elementtree_parse_file(self.directory + "/" + item)
        members = []

        parent = self.strip_extension(item)

        for item in self.get_item_elements(root):
            members.append(self.get_item_name(item, parent))

        return members

    def check_delete_excludes(self, item):
        return False

    def get_item_elements(self, root):
        return root.findall(self.item_xpath, self.namespaces)

    def get_name_elements(self, item):
        return item.findall(self.name_xpath, self.namespaces)

    def get_item_name(self, item, parent):
        """Returns the value of the first name element found inside of element"""
        names = self.get_name_elements(item)
        if not names:
            raise MissingNameElementError

        name = names[0].text
        prefix = self.item_name_prefix(parent)
        if prefix:
            name = prefix + name

        return name

    def item_name_prefix(self, parent):
        return parent + "."

    def _strip_component(self, item, component_list):
        parent = self.strip_extension(item)
        package_tree = elementtree_parse_file(
            os.path.join(self.directory, item), namespace=self.namespaces["sf"]
        )
        root = package_tree.getroot()
        for element in self.get_item_elements(root):
            component_name = (
                self.item_name_prefix(parent) + self.get_name_elements(element)[0].text
            )
            if component_name not in component_list:
                self.logger.info(
                    f"Deleting component {component_name} of type {self.metadata_type}"
                )
                root.remove(element)
        package_tree.write(
            os.path.join(self.directory, item), encoding="UTF-8", xml_declaration=True
        )


# TYPE SPECIFIC PARSERS


class CustomLabelsParser(MetadataXmlElementParser):
    def item_name_prefix(self, parent):
        return ""


class CustomObjectParser(MetadataFilenameParser):
    def _parse_item(self, item):
        members = []

        # Skip namespaced custom objects
        if len(item.split("__")) > 2:
            return members

        # Skip standard objects
        if (
            not item.endswith("__c.object")
            and not item.endswith("__mdt.object")
            and not item.endswith("__e.object")
            and not item.endswith("__b.object")
        ):
            return members

        members.append(self.strip_extension(item))
        return members

    def _strip_component(self, item, component_list):
        # Skip namespaced custom objects
        if len(item.split("__")) > 2:
            return

        # Skip standard objects
        if (
            not item.endswith("__c.object")
            and not item.endswith("__mdt.object")
            and not item.endswith("__e.object")
            and not item.endswith("__b.object")
        ):
            return

        if self.strip_extension(item) not in component_list:
            self.logger.info(
                f"Deleting component {self.strip_extension(item)} of type {self.metadata_type}"
            )
            path = os.path.join(self.directory, item)
            os.remove(path)


class RecordTypeParser(MetadataXmlElementParser):
    def check_delete_excludes(self, item):
        if self.delete:
            return True


class BusinessProcessParser(MetadataXmlElementParser):
    def check_delete_excludes(self, item):
        if self.delete:
            return True


class BundleParser(BaseMetadataParser):
    def _parse_item(self, item):
        members = []
        path = self.directory + "/" + item

        # Skip non-directories
        if not os.path.isdir(path):
            return members

        # item is a directory; add directory to members and ignore processing directory's files
        members.append(item)

        return members

    def _strip_component(self, item, component_list):
        path = os.path.join(self.directory, item)
        # Check if it is a directory and take action
        if os.path.isdir(path):
            if item not in component_list:
                self.logger.info(
                    f"Deleting component {item} of type {self.metadata_type}"
                )
                shutil.rmtree(path)


class LWCBundleParser(BaseMetadataParser):
    def _parse_item(self, item):
        members = []
        path = self.directory + "/" + item

        # Skip non-directories
        if not os.path.isdir(path) or item.startswith("__"):
            return members

        # item is a directory; add directory to members and ignore processing directory's files
        members.append(item)

        return members

    def _strip_component(self, item, component_list):
        path = os.path.join(self.directory, item)
        # Check if it is a directory and take action
        if os.path.isdir(path):
            if item not in component_list:
                self.logger.info(
                    f"Deleting component {item} of type {self.metadata_type}"
                )
                shutil.rmtree(path)


class DocumentParser(MetadataFolderParser):
    def _parse_subitem(self, item, subitem):
        return [item + "/" + subitem]

    def _strip_subitem(self, item, subitem, component_list):
        if item + "/" + subitem not in component_list:
            path = os.path.join(self.directory, item, subitem)
            self.logger.info(
                f"Deleting component {item}/{subitem} of type {self.metadata_type}"
            )
            os.remove(path)
            if os.path.exists(path + "-meta.xml"):
                os.remove(path + "-meta.xml")


class UpdatePackageXml(BaseTask):
    task_options = {
        "path": {
            "description": "The path to a folder of metadata to build the package.xml from",
            "required": True,
        },
        "output": {"description": "The output file, defaults to <path>/package.xml"},
        "package_name": {
            "description": "If set, overrides the package name inserted into the <fullName> element"
        },
        "managed": {
            "description": "If True, generate a package.xml for deployment to the managed package packaging org"
        },
        "delete": {
            "description": "If True, generate a package.xml for use as a destructiveChanges.xml file for deleting metadata"
        },
        "install_class": {
            "description": "Specify post install class file to be used. Defaults to what is set in project config"
        },
        "uninstall_class": {
            "description": "Specify post uninstall class file to be used. Defaults to what is set in project config"
        },
    }

    def _init_options(self, kwargs):
        super(UpdatePackageXml, self)._init_options(kwargs)
        self.options["managed"] = self.options.get("managed") in (True, "True", "true")

    def _init_task(self):
        package_name = self.options.get("package_name")
        if not package_name:
            if self.options.get("managed"):
                package_name = self.project_config.project__package__name_managed
            if not package_name:
                package_name = self.project_config.project__package__name

        self.package_xml = PackageXmlGenerator(
            directory=self.options.get("path"),
            api_version=self.project_config.project__package__api_version,
            package_name=package_name,
            managed=self.options.get("managed", False),
            delete=self.options.get("delete", False),
            install_class=self.options.get(
                "install_class", self.project_config.project__package__install_class
            ),
            uninstall_class=self.options.get(
                "uninstall_class", self.project_config.project__package__uninstall_class
            ),
        )

    def _run_task(self):
        output = self.options.get(
            "output", "{}/package.xml".format(self.options.get("path"))
        )
        self.logger.info(
            "Generating {} from metadata in {}".format(output, self.options.get("path"))
        )
        package_xml = self.package_xml()
        with open(self.options.get("output", output), mode="w", encoding="utf-8") as f:
            f.write(package_xml)


class RemoveSourceComponents:
    def __init__(self, directory, package_xml, api_version, logger):
        self.directory = directory
        self.package_xml = package_xml
        self.api_version = api_version
        self.logger = logger

    def __call__(self):
        xml_map = self.generate_package_xml_map()
        self.folder_parser = PackageXmlGenerator(
            directory=self.directory, api_version=self.api_version, logger=self.logger
        )
        self.folder_parser.parse_types()
        self.folder_parser.types.sort(key=lambda x: x.metadata_type.upper())
        for parse_type in self.folder_parser.types:
            parse_type.strip_folder(
                xml_map[parse_type.metadata_type]
                if parse_type.metadata_type in xml_map
                else []
            )

    def generate_package_xml_map(self) -> dict:
        package = metadata_tree.parse(self.package_xml)
        xml_map = {}
        for type in package.types:
            members = []
            try:
                for member in type.members:
                    members.append(member.text)
            except AttributeError:  # Exception if there are no members for a type
                pass
            xml_map[type["name"].text] = members
        return xml_map


class RemoveUnwantedComponents(BaseTask):
    task_options = {
        "path": {
            "description": "The path to a folder of metadata to strip the components",
            "required": True,
        },
        "package_xml": {
            "description": "The path to package xml file to refer",
            "required": True,
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.run_class = RemoveSourceComponents(
            directory=self.options.get("path"),
            package_xml=self.options.get("package_xml"),
            api_version=self.project_config.project__package__api_version,
            logger=self.logger,
        )

    def _run_task(self):
        self.logger.info(
            "Removing unwanted components from " + self.options.get("path")
        )
        self.run_class()



import json
import typing as T
from contextlib import contextmanager
from functools import lru_cache
from io import StringIO
from pathlib import Path
from unittest import mock

import pytest
from sqlalchemy import create_engine

from cumulusci.core.config import OrgConfig
from cumulusci.salesforce_api.org_schema import Filters, get_org_schema
from cumulusci.tasks.bulkdata.extract_dataset_utils.extract_yml import ExtractRulesFile
from cumulusci.tasks.bulkdata.extract_dataset_utils.synthesize_extract_declarations import (
    flatten_declarations,
)
from cumulusci.tasks.bulkdata.step import DataApi
from cumulusci.tests.util import read_mock
from cumulusci.utils.tests.test_org_schema import FakeSF


class TestSynthesizeExtractDeclarations:
    def test_synthesize_extract_declarations(self, org_config):
        declarations = """
            extract:
                    OBJECTS(ALL):
                        fields:
                            FIELDS(REQUIRED)
        """
        object_counts = {"Account": 10, "Contact": 0, "Case": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema, ())

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": "Name != 'Sample Account for Entitlements'",
                        "fields_": mock.ANY,
                        "api": DataApi.SMART,
                        "sf_object": "Account",
                    },
                    {
                        "where": None,
                        "fields_": mock.ANY,
                        "api": DataApi.SMART,
                        "sf_object": "Case",
                    },
                )
            )

    def test_find_custom_objects(self, org_config):
        declarations = """
            extract:
                    OBJECTS(CUSTOM):
                        fields:
                            FIELDS(ALL)
        """
        object_counts = {"Account": 10, "Contact": 0, "Case": 5, "Custom__c": 10}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
            describe_for("Custom__c"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": None,
                        "fields_": mock.ANY,
                        "api": DataApi.SMART,
                        "sf_object": "Custom__c",
                    },
                )
            )

    def test_find_standard_objects(self, org_config):
        declarations = """
            extract:
                    OBJECTS(STANDARD):
                        fields:
                            FIELDS(REQUIRED)
        """
        object_counts = {"Account": 10, "Contact": 0, "Case": 5, "Custom__c": 10}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
            describe_for("Custom__c"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": "Name != 'Sample Account for Entitlements'",
                        "fields_": mock.ANY,
                        "api": DataApi.SMART,
                        "sf_object": "Account",
                    },
                    {
                        "where": None,
                        "fields_": mock.ANY,
                        "api": DataApi.SMART,
                        "sf_object": "Case",
                    },
                )
            )

    def test_filter_objects(self, org_config):
        declarations = """
            extract:
                    OBJECTS(STANDARD):
                        fields:
                            FIELDS(REQUIRED)
        """
        object_counts = {"Account": 10, "Contact": 0, "Case": 5, "Custom__c": 10}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Case"),
            describe_for("Custom__c"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            filters=[],
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema, ("Account"))

            assert tuple(decl.dict() for decl in decls) == tuple(
                (  # No Account
                    {
                        "where": None,
                        "fields_": mock.ANY,
                        "api": DataApi.SMART,
                        "sf_object": "Case",
                    },
                )
            )

    def test_synthesize_extract_declarations__custom(self, org_config):
        declarations = """
            extract:
                    OBJECTS(CUSTOM):
                        fields:
                            FIELDS(ALL)
        """
        object_counts = {"Account": 0, "Contact": 2, "Custom__c": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Custom__c"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": None,
                        "fields_": mock.ANY,
                        "api": DataApi.SMART,
                        "sf_object": "Custom__c",
                    },
                )
            )

    def test_synthesize_extract_declarations__custom_fields(self, org_config):
        declarations = """
            extract:
                    OBJECTS(CUSTOM):
                        fields:
                            FIELDS(custom)
        """
        object_counts = {"Account": 0, "Contact": 2, "Custom__c": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Custom__c"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": None,
                        "fields_": ["CustomField__c"],
                        "api": DataApi.SMART,
                        "sf_object": "Custom__c",
                    },
                )
            )

    def test_synthesize_extract_declarations__standard_fields(self, org_config):
        declarations = """
            extract:
                    OBJECTS(CUSTOM):
                        fields:
                            FIELDS(standard)
        """
        object_counts = {"Account": 0, "Contact": 2, "Custom__c": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Custom__c"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": None,
                        "fields_": ["Name"],
                        "api": DataApi.SMART,
                        "sf_object": "Custom__c",
                    },
                )
            )

    def test_required_lookups__pulled_in(self, org_config):
        """Bringing in the AccountId should force Account to come in.

        Including all Account/Contact required fields."""
        declarations = """
            extract:
                Contact:
                    fields:
                        AccountId
        """
        object_counts = {"Account": 3, "Contact": 2, "Custom__c": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Custom__c"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": mock.ANY,
                        "fields_": ["AccountId", "LastName"],
                        "api": DataApi.SMART,
                        "sf_object": "Contact",
                    },
                    {
                        "where": mock.ANY,
                        "fields_": [
                            "Name",
                        ],
                        "api": DataApi.SMART,
                        "sf_object": "Account",
                    },
                )
            )

    def test_required_lookups__pulled_in__polymorphic_lookups(self, org_config):
        """Bringing in the WhoId for sobject Event should force Contact
        and Lead to come in.

        Including all Lead/Contact/Event required fields."""
        declarations = """
            extract:
                Event:
                    fields:
                        WhoId
        """
        object_counts = {
            "Account": 3,
            "Contact": 2,
            "Custom__c": 5,
            "Lead": 2,
            "Event": 1,
        }
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Custom__c"),
            describe_for("Event"),
            describe_for("Lead"),
        )
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)

            assert tuple(decl.dict() for decl in decls) == tuple(
                (
                    {
                        "where": mock.ANY,
                        "fields_": ["WhoId"],
                        "api": DataApi.SMART,
                        "sf_object": "Event",
                    },
                    {
                        "where": mock.ANY,
                        "fields_": ["LastName"],
                        "api": DataApi.SMART,
                        "sf_object": "Contact",
                    },
                    {
                        "where": mock.ANY,
                        "fields_": ["Company", "LastName"],
                        "api": DataApi.SMART,
                        "sf_object": "Lead",
                    },
                )
            )

    def test_parse_real_file(self, cumulusci_test_repo_root, org_config):
        declarations = ExtractRulesFile.parse_extract(
            cumulusci_test_repo_root / "datasets/test_minimal.extract.yml"
        )
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Opportunity"),
            describe_for("Custom__c"),
        )
        object_counts = {
            "Account": 100,
            "Opportunity": 10,
            "Custom__c": 5,
            "Contact": 0,
        }
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)
            decls = {decl.sf_object: decl for decl in decls}
            assert decls["Opportunity"].fields == [
                "Name",
                "ContactId",
                "AccountId",
                "CloseDate",  # pull these in because they required
                "StageName",
            ]
            assert "Account" in decls
            assert "Contact" in decls  # pulled in because of ContactId
            assert "Custom__c" in decls
            assert set(decls["Custom__c"].fields) == set(
                schema["Custom__c"].fields.keys()
            ) - set(["Id"])

    def test_synthesize_extract_declarations__where_clause(self, org_config):
        declarations = """
            extract:
                Contact:
                    fields:
                        FIELDS(REQUIRED)
                Account:
                    where: Name Like '%Foo%'
                    fields:
                        - Description

        """
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        object_counts = {"Account": 2, "Contact": 2, "Custom__c": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Custom__c"),
        )
        with _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)
            decls = {decl.sf_object: decl for decl in decls}
            assert "Account" in decls
            assert "Contact" in decls
            assert "Custom__c" not in decls

            assert set(decls["Account"].fields) == set(["Name", "Description"])
            assert decls["Contact"].fields == ["LastName"], decls["Contact"].fields

    def test_synthesize_fields(self, sf, org_config):
        declarations = """
            extract:
                Opportunity:
                    fields:
                        - FIELDS(STANDARD)
                Account:
                    fields:
                        - FIELDS(CUSTOM)
                Contact:
                    fields:
                        - FIELDS(ALL)
                Custom__c:
                    fields:
                        - FIELDS(CUSTOM)
        """
        declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
        object_counts = {"Account": 2, "Contact": 2, "Custom__c": 5}
        object_describes = [describe_for(obj) for obj in object_counts.keys()]
        with _fake_get_org_schema(
            org_config,
            object_describes,
            object_counts,
            include_counts=True,
        ) as schema:
            decls = flatten_declarations(declarations.values(), schema)
            decls = {decl.sf_object: decl for decl in decls}
            assert "Account" in decls
            assert "Contact" in decls
            assert "Entitlement" not in decls
            assert "Opportunity" not in decls  # not populated
            assert "Name" not in decls["Custom__c"].fields
            assert "Id" not in decls["Custom__c"].fields
            assert "Name" in decls["Account"].fields  # because required
            assert "BillingCountry" not in decls["Account"].fields  # not required
            assert "CustomField__c" in decls["Custom__c"].fields

    @pytest.mark.needs_org()
    @pytest.mark.slow()
    def test_find_standard_objects__integration_tests(self, sf, org_config):
        declarations = """
            extract:
                    OBJECTS(STANDARD):
                        fields:
                            FIELDS(REQUIRED)
        """
        with get_org_schema(
            sf,
            org_config,
            include_counts=True,
            filters=[Filters.createable, Filters.extractable, Filters.populated],
        ) as schema:
            declarations = ExtractRulesFile.parse_extract(StringIO(declarations))
            decls = flatten_declarations(declarations.values(), schema)
            decls = {decl.sf_object: decl for decl in decls}
            assert "WorkBadgeDefinition" in decls
            # HEY NOW!
            assert "You\\'re a RockStar!" in str(decls["WorkBadgeDefinition"].where)
            if "Opportunity" in decls:
                assert "IsPrivate" not in decls["Opportunity"].fields, decls.keys()


@lru_cache(maxsize=None)
def describe_for(sobject: str):
    return json.loads(read_mock(sobject))


@contextmanager
def _fake_get_org_schema(
    org_config: OrgConfig,
    org_describes: T.Sequence[dict],
    object_counts: T.Dict[str, int],
    **kwargs,
):
    with mock.patch(
        "cumulusci.salesforce_api.org_schema.count_sobjects",
        lambda *args: (
            object_counts,
            [],
            [],
        ),
    ), mock.patch(
        "cumulusci.salesforce_api.org_schema.ZippableTempDb", FakeZippableTempDb
    ), mock.patch(
        "cumulusci.salesforce_api.org_schema.deep_describe",
        return_value=((desc, "Sat, 1 Jan 2000 00:00:01 GMT") for desc in org_describes),
    ), get_org_schema(
        FakeSF(), org_config, **kwargs
    ) as schema:
        yield schema


class FakeZippableTempDb:
    "Fast no-IO database for testing"

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        return ""

    def zip_database(self, target_path):
        pass

    def unzip_database(self, zipped_db):
        pass

    def clear(self):
        pass

    def create_engine(self):
        self.engine = create_engine("sqlite:///")
        return self.engine


@contextmanager
def faketempdb():
    yield Path("")


# TODO: Decide upon what happens in this cases:
#       OBJECTS(STANDARD)
#          fields:
#           - X
#       Account:
#          fields:
#           - Y
#    Do you get X and Y? (Union semantics?) Or just Y (override semantics)
# Consider all kinds of conflicting/complementary declarations.



import os
import sys
from datetime import date, timedelta
from unittest import mock

import click
import pytest

import cumulusci
from cumulusci.cli.runtime import CliRuntime
from cumulusci.core.config import OrgConfig
from cumulusci.core.exceptions import ConfigError, OrgNotFound


class TestCliRuntime:
    key = "1234567890abcdef"

    def setup_method(self):
        os.chdir(os.path.dirname(cumulusci.__file__))
        self.environ_mock = mock.patch.dict(os.environ, {"CUMULUSCI_KEY": self.key})
        self.environ_mock.start()

    def teardown_method(self):
        self.environ_mock.stop()

    def test_init(self):
        config = CliRuntime()

        for key in {"cumulusci", "tasks", "flows", "services", "orgs", "project"}:
            assert key in config.universal_config.config
        assert config.project_config.project__name == "CumulusCI"
        for key in {"services", "orgs"}:
            assert key in config.keychain.config
        assert config.project_config.repo_root in sys.path

    @mock.patch("cumulusci.cli.runtime.CliRuntime._load_project_config")
    def test_load_project_config_error(self, load_proj_cfg_mock):
        load_proj_cfg_mock.side_effect = ConfigError

        with pytest.raises(click.UsageError):
            CliRuntime()

    @mock.patch("cumulusci.cli.runtime.keyring")
    def test_get_keychain_key__migrates_from_env_to_keyring(self, keyring):
        keyring.get_password.return_value = None

        config = CliRuntime()
        assert self.key == config.keychain.key
        keyring.set_password.assert_called_once_with(
            "cumulusci", "CUMULUSCI_KEY", self.key
        )

    @mock.patch("cumulusci.cli.runtime.keyring")
    def test_get_keychain_key__env_takes_precedence(self, keyring):
        if os.environ.get("CUMULUSCI_KEYCHAIN_CLASS"):
            del os.environ["CUMULUSCI_KEYCHAIN_CLASS"]
        keyring.get_password.return_value = "overridden"

        config = CliRuntime()
        assert self.key == config.keychain.key

    @mock.patch("cumulusci.cli.runtime.keyring")
    def test_get_keychain_key__generates_key(self, keyring):
        with mock.patch.dict(os.environ):
            del os.environ["CUMULUSCI_KEY"]
            if os.environ.get("CUMULUSCI_KEYCHAIN_CLASS"):
                del os.environ["CUMULUSCI_KEYCHAIN_CLASS"]
            keyring.get_password.return_value = None

            config = CliRuntime()
        assert self.key != config.keychain.key
        assert 16 == len(config.keychain.key)

    @mock.patch("cumulusci.cli.runtime.keyring")
    def test_get_keychain_key__warns_if_generated_key_cannot_be_stored(self, keyring):
        with mock.patch.dict(os.environ):
            del os.environ["CUMULUSCI_KEY"]
            keyring.get_password.side_effect = Exception

            runtime = CliRuntime()
            assert runtime.keychain_key is None

    def test_get_org(self):
        config = CliRuntime()
        config.keychain = mock.Mock()
        config.keychain.get_org.return_value = org_config = OrgConfig({}, "test")

        org_name, org_config_result = config.get_org("test")
        assert org_name == "test"
        assert org_config is org_config_result

    def test_get_org_default(self):
        config = CliRuntime()
        config.keychain = mock.Mock()
        org_config = OrgConfig({}, "test")
        config.keychain.get_default_org.return_value = ("test", org_config)

        org_name, org_config_result = config.get_org()
        assert org_name == "test"
        assert org_config is org_config_result

    def test_get_org_missing(self):
        config = CliRuntime()
        config.keychain = mock.Mock()
        config.keychain.get_org.return_value = None

        with pytest.raises(click.UsageError):
            org_name, org_config_result = config.get_org("test", fail_if_missing=True)

    def test_check_org_expired(self):
        config = CliRuntime()
        config.keychain = mock.Mock()
        org_config = OrgConfig(
            {
                "scratch": True,
                "date_created": date.today() - timedelta(days=2),
                "expired": True,
            },
            "test",
        )

        config.check_org_expired("test", org_config)
        config.keychain.create_scratch_org.assert_called_once()

    def test_check_org_overwrite_not_found(self):
        config = CliRuntime()
        config.keychain.get_org = mock.Mock(side_effect=OrgNotFound)

        assert config.check_org_overwrite("test")

    def test_check_org_overwrite_scratch_exists(self):
        config = CliRuntime()
        config.keychain.get_org = mock.Mock(
            return_value=OrgConfig({"scratch": True, "created": True}, "test")
        )

        with pytest.raises(click.ClickException):
            config.check_org_overwrite("test")

    def test_check_org_overwrite_non_scratch_exists(self):
        config = CliRuntime()
        config.keychain.get_org = mock.Mock(
            return_value=OrgConfig({"scratch": False}, "test")
        )

        with pytest.raises(click.ClickException):
            config.check_org_overwrite("test")

    def test_check_cumulusci_version(self):
        config = CliRuntime()
        config.project_config.minimum_cumulusci_version = "999"

        with pytest.raises(click.UsageError):
            config.check_cumulusci_version()

    @mock.patch("cumulusci.cli.runtime.call")
    @mock.patch("cumulusci.cli.runtime.click.echo")
    @mock.patch("sys.platform", "darwin")
    def test_alert_osx(self, echo_mock, shell_mock):
        config = CliRuntime()

        config.alert("hello")
        echo_mock.assert_called_once()
        shell_mock.assert_called_once()
        assert "osascript" in shell_mock.call_args[0][0]

    @mock.patch("cumulusci.cli.runtime.call")
    @mock.patch("cumulusci.cli.runtime.click.echo")
    @mock.patch("sys.platform", "linux2")
    def test_alert_linux(self, echo_mock, shell_mock):
        config = CliRuntime()

        config.alert("hello")
        echo_mock.assert_called_once()
        shell_mock.assert_called_once()
        assert "notify-send" in shell_mock.call_args[0][0]

    @mock.patch("cumulusci.cli.runtime.call")
    @mock.patch("cumulusci.cli.runtime.click.echo")
    @mock.patch("sys.platform", "darwin")
    def test_alert__disabled(self, echo_mock, shell_mock):
        config = CliRuntime()
        config.project_config.dev_config__no_alert = True

        config.alert("hello")
        echo_mock.assert_not_called()
        shell_mock.assert_not_called()

    @mock.patch("cumulusci.cli.runtime.call")
    @mock.patch("cumulusci.cli.runtime.click.echo")
    @mock.patch("sys.platform", "darwin")
    def test_alert__os_error(self, echo_mock, shell_mock):
        shell_mock.side_effect = OSError
        config = CliRuntime()
        config.alert("hello")
        echo_mock.assert_called_once()
        shell_mock.assert_called_once()



from cumulusci.core import exceptions


class TestExceptions:
    def test_ApexCompilationException_str(self):
        err = exceptions.ApexCompilationException(42, "Too many braces")
        assert str(err) == "Apex compilation failed on line 42: Too many braces"

    def test_ApexException(self):
        err = exceptions.ApexException("Things got real", "Anon line 1")
        assert str(err) == "Apex error: Things got real\n  Stacktrace:\n  Anon line 1"



import json
from typing import Dict, List, Union

import salesforce_bulk
from requests.exceptions import ConnectionError
from simple_salesforce.exceptions import SalesforceGeneralError

from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask
from cumulusci.utils.parallel.queries_in_parallel.run_queries_in_parallel import (
    RunParallelQueries,
)


class MetadataInfo:
    def __init__(
        self, columns: List[str], table_name: str, package_xml_name: str, id_field: str
    ):
        self.columns = columns
        self.table_name = table_name
        self.package_xml_name = package_xml_name
        self.id_field = id_field


APEXCLASS = MetadataInfo(
    columns=["Name", "NamespacePrefix"],
    table_name="ApexClass",
    package_xml_name="ApexClass",
    id_field="Id",
)
APEXPAGE = MetadataInfo(
    columns=["Name", "NamespacePrefix"],
    table_name="ApexPage",
    package_xml_name="ApexPage",
    id_field="Id",
)
CUSTOMPERMISSION = MetadataInfo(
    columns=["DeveloperName", "NamespacePrefix"],
    table_name="CustomPermission",
    package_xml_name="CustomPermission",
    id_field="Id",
)
TABSET = MetadataInfo(
    columns=["Name", "NamespacePrefix"],
    table_name="AppMenuItem",
    package_xml_name="CustomApplication",
    id_field="ApplicationId",
)
CONNECTEDAPPLICATION = MetadataInfo(
    columns=["Name", "NamespacePrefix"],
    table_name="AppMenuItem",
    package_xml_name="CustomApplication",
    id_field="ApplicationId",
)
EXTERNALDATASOURCE = MetadataInfo(
    columns=["DeveloperName", "NamespacePrefix"],
    table_name="ExternalDataSource",
    package_xml_name="ExternalDataSource",
    id_field="Id",
)
FLOWDEFINITION = MetadataInfo(
    columns=["ApiName"],
    table_name="FlowDefinitionView",
    package_xml_name="Flow",
    id_field="Id",
)

SETUP_ENTITY_TYPES = {
    "ApexClass": APEXCLASS,
    "ApexPage": APEXPAGE,
    "CustomPermission": CUSTOMPERMISSION,
    "TabSet": TABSET,
    "ConnectedApplication": CONNECTEDAPPLICATION,
    "ExternalDataSource": EXTERNALDATASOURCE,
    "FlowDefinition": FLOWDEFINITION,
}
SETUP_ENTITY_QUERY_NAME = "setupEntityAccess"

SOBJECT_TYPE = "CustomObject"
SOBJECT_QUERY_NAME = "sObject"

CUSTOM_TAB_TYPE = "CustomTab"
CUSTOM_TAB_QUERY_NAME = "customTab"

PROFILE_FLOW_QUERY_NAME = "profileFlow"


class RetrieveProfileApi(BaseSalesforceApiTask):
    def _init_task(self):
        super(RetrieveProfileApi, self)._init_task()
        self.api_version = self.project_config.config["project"]["package"][
            "api_version"
        ]

    def _retrieve_existing_profiles(self, profiles: List[str]):
        query = self._build_query(["Name"], "Profile", {"Name": profiles})
        result = self._run_query(query)

        existing_profiles = []
        for data in result["records"]:
            existing_profiles.append(data["Name"])

        # Since System Administrator is named Admin in Metadata API
        if "Admin" in profiles:
            existing_profiles.extend(["Admin", "System Administrator"])

        return existing_profiles

    def _run_query(self, query):
        try:
            result = self.sf.query(query)
        except (SalesforceGeneralError, ConnectionError):
            result = self._run_bulk_query(query)
        return result

    def _run_bulk_query(self, query):
        table_name = self._extract_table_name_from_query(query)
        job = self.bulk.create_query_job(table_name, contentType="JSON")

        batch = self.bulk.query(job, query)
        self.bulk.wait_for_batch(job, batch)
        self.bulk.close_job(job)
        results = self.bulk.get_all_results_for_query_batch(batch)
        for result in results:
            result = json.load(salesforce_bulk.util.IteratorBytesIO(result))
            return {"records": result}

    def _extract_table_name_from_query(self, query):
        from_index = query.upper().find("FROM")
        if from_index != -1:
            query = query[from_index + len("FROM") :].strip()
            table_name = query.split()[0]
            return table_name
        else:
            raise ValueError("Invalid query format. 'FROM' clause not found.")

    def _build_query(
        self,
        columns: List[str],
        table_name: str,
        where: Dict[str, Union[str, List[str]]] = None,
    ) -> str:
        if not columns:
            raise ValueError("Columns list cannot be empty")
        if not table_name:
            raise ValueError("Table name cannot be empty")

        select_clause = ", ".join(columns)

        where_clause = ""
        if where:
            where_list = []
            for key, value in where.items():
                if isinstance(value, list):
                    value = ", ".join([f"'{item}'" for item in value])
                    condition = f"{key} IN ({value})"
                else:
                    condition = f"{key} = '{value}'"
                where_list.append(condition)
            where_clause = "WHERE " + " AND ".join(where_list)

        query = f"SELECT {select_clause} FROM {table_name} {where_clause}".strip()
        return query

    def _queries_retrieve_permissions(self, profiles: List[str]):
        queries = {}

        # Setup Entity Access query
        setupEntityAccess_query = self._build_query(
            ["SetupEntityId", "SetupEntityType"],
            "SetupEntityAccess",
            {"Parent.Profile.Name": profiles},
        )
        queries[SETUP_ENTITY_QUERY_NAME] = setupEntityAccess_query

        # sObject query
        sObject_query = self._build_query(
            ["SObjectType"], "ObjectPermissions", {"Parent.Profile.Name": profiles}
        )
        queries[SOBJECT_QUERY_NAME] = sObject_query

        # Custom Tab query
        customTab_query = self._build_query(
            ["Name"], "PermissionSetTabSetting", {"Parent.Profile.Name": profiles}
        )
        queries[CUSTOM_TAB_QUERY_NAME] = customTab_query

        # Matching Profile Name and Flow query
        profileFlow_query = self._build_query(
            ["SetupEntityId", "Parent.Profile.Name"],
            "SetupEntityAccess",
            {"Parent.Profile.Name": profiles, "SetupEntityType": "FlowDefinition"},
        )
        queries[PROFILE_FLOW_QUERY_NAME] = profileFlow_query

        return queries

    def _process_all_results(self, result: dict):
        permissionable_entities = {}
        entities, result_setupEntityAccess = self._process_setupEntityAccess_results(
            result[SETUP_ENTITY_QUERY_NAME]
        )
        permissionable_entities.update(entities)
        permissionable_entities.update(
            self._process_sObject_results(result[SOBJECT_QUERY_NAME])
        )
        permissionable_entities.update(
            self._process_customTab_results(result[CUSTOM_TAB_QUERY_NAME])
        )

        # Process the profile and flows
        profile_flow = self._match_profiles_and_flows(
            result[PROFILE_FLOW_QUERY_NAME],
            result_setupEntityAccess.get("FlowDefinition", []),
        )

        return permissionable_entities, profile_flow

    # Retrieve all the permissionable entitites for a set of profiles
    def _retrieve_permissionable_entities(self, profiles: List[str]):
        # Logs
        self.logger.info("Querying for all permissionable entities:")
        self.logger.info("Pending")

        # Run all queries
        result = RunParallelQueries._run_queries_in_parallel(
            self._queries_retrieve_permissions(profiles), self._run_query
        )

        # Process the results
        permissionable_entities, profile_flow = self._process_all_results(result)

        # Logs
        self.logger.info("[Done]\n")

        return permissionable_entities, profile_flow

    def _process_setupEntityAccess_results(self, result_list: List[dict]):
        setupEntityAccess_dict = {
            entity_type: [] for entity_type in SETUP_ENTITY_TYPES.keys()
        }

        for data in result_list:
            entity_type = data["SetupEntityType"]
            entity_id = data["SetupEntityId"]
            if entity_type in SETUP_ENTITY_TYPES.keys():
                setupEntityAccess_dict[entity_type].append(entity_id)

        queries = {}
        for entity_type, query_values in SETUP_ENTITY_TYPES.items():
            if query_values and len(setupEntityAccess_dict[entity_type]) != 0:
                queries[entity_type] = self._build_query(
                    query_values.columns,
                    query_values.table_name,
                    {query_values.id_field: setupEntityAccess_dict[entity_type]},
                )

        result = RunParallelQueries._run_queries_in_parallel(queries, self._run_query)

        extracted_values = {}
        for entity_type, data in SETUP_ENTITY_TYPES.items():
            if entity_type in result and data:
                extracted_values.setdefault(data.package_xml_name, [])
                for item in result[entity_type]:
                    if (
                        "NamespacePrefix" in item
                        and item["NamespacePrefix"] is not None
                    ):
                        extracted_values[data.package_xml_name].append(
                            f'{item["NamespacePrefix"]}__{item[data.columns[0]]}'
                        )
                    else:
                        extracted_values[data.package_xml_name].append(
                            item[data.columns[0]]
                        )

        return extracted_values, result

    def _process_sObject_results(self, result_list: List[dict]):
        sObject_list = [data["SobjectType"] for data in result_list]
        return {SOBJECT_TYPE: sObject_list}

    def _process_customTab_results(self, result_list: List[dict]):
        customTab_list = [data["Name"] for data in result_list]
        return {CUSTOM_TAB_TYPE: customTab_list}

    def _match_profiles_and_flows(
        self, result_profiles: List[dict], result_flows: List[dict]
    ):
        profile_mapping = {}
        for item in result_profiles:
            setup_entity_id = item["SetupEntityId"]
            profile_name = item["Parent"]["Profile"]["Name"]
            profile_mapping.setdefault(setup_entity_id, []).append(profile_name)

        result_dict = {}
        for flow in result_flows:
            setup_entity_id = flow["attributes"]["url"].split("/")[-1]
            for profile_name in profile_mapping.get(setup_entity_id, []):
                result_dict.setdefault(profile_name, []).append(flow["ApiName"])

        return result_dict



from cumulusci.core.tasks import BaseTask
from cumulusci.utils.xml import lxml_parse_string


class BaseMarketingCloudTask(BaseTask):
    """Base task for interacting with Marketing Cloud

    For API calls to a MC tenant, you can get a fresh
    access token via the marketing cloud config like so:

    self.mc_config.access_token
    """

    salesforce_task = False

    def _init_task(self):
        super()._init_task()
        self.mc_config = self.project_config.keychain.get_service("marketing_cloud")

    def _check_soap_response(self, response):
        """Make sure the response indicates success."""
        response.raise_for_status()
        root = lxml_parse_string(response.content)
        status_code = root.find(
            ".//{http://exacttarget.com/wsdl/partnerAPI}StatusCode"
        ).text
        status_message = root.find(
            ".//{http://exacttarget.com/wsdl/partnerAPI}StatusMessage"
        ).text
        if status_code != "OK":
            raise Exception(
                f"Error from Marketing Cloud: {status_message}\n\nFull response text: {response.text}"
            )

    def get_mc_stack_key(self) -> str:
        """Return the stack_key associated with this tasks's marketing_cloud service"""
        user_info_payload = self.mc_config.get_user_info()
        return user_info_payload["organization"]["stack_key"]



from pathlib import Path
from unittest import mock

from robot.libraries.BuiltIn import RobotNotRunningError

from cumulusci.robotframework.Salesforce import Salesforce


# FIXME: we shouldn't have to tweak these tests for every
# version. The tests should be smarter.
class TestLocators:
    @mock.patch(
        "cumulusci.robotframework.SalesforceAPI.SalesforceAPI.get_latest_api_version"
    )
    def test_locators_in_robot_context(self, get_latest_api_version):
        """Verify we can get locators for the current org api version"""
        get_latest_api_version.return_value = 57.0

        # This instantiates the robot library, mimicking a robot library import.
        # We've mocked out the code that would otherwise throw an error since
        # we're not running in the context of a robot test. The library should
        # return the latest version of the locators.
        sf = Salesforce()

        expected = "cumulusci.robotframework.locators_57"
        actual = sf.locators_module.__name__
        message = "expected to load '{}', actually loaded '{}'".format(expected, actual)
        assert expected == actual, message

    @mock.patch(
        "robot.libraries.BuiltIn.BuiltIn.get_library_instance",
        side_effect=RobotNotRunningError(),
    )
    def test_locators_outside_robot_context(self, builtin_mock):
        """Verify that we get the latest locators if not running in the context of a robot test"""

        # This instantiates the robot library, mimicing a robot library import
        # however, because we've mocked get_library_instance to throw an error,
        # we expect the library to still be instantiated, but with the latest
        # version of the locators.

        sf = Salesforce()

        locator_folder = Path("./cumulusci/robotframework")
        locator_modules = sorted(locator_folder.glob("locators_[0-9][0-9].py"))
        expected = f"cumulusci.robotframework.{locator_modules[-1].stem}"

        actual = sf.locators_module.__name__
        message = "expected to load '{}', actually loaded '{}'".format(expected, actual)
        assert expected == actual, message

    def test_locators_57(self):
        """Verify that locators_57 is a superset of the locators_56

        This test is far from perfect, but it should at least flag a
        catastrophic error in how locators for a version that augments
        the locators from previous versions.

        Note: this test assumes that locators_56 doesn't delete any of the
        keys from 54.

        """
        import cumulusci.robotframework.locators_56 as locators_56
        import cumulusci.robotframework.locators_57 as locators_57

        keys_56 = set(locators_56.lex_locators)
        keys_57 = set(locators_57.lex_locators)

        assert id(locators_56.lex_locators) != id(
            locators_57.lex_locators
        ), "locators_56.lex_locators and locators_57.lex_locators are the same object"
        assert len(keys_56) > 0
        assert keys_57.issubset(keys_56)



""" Tasks are the basic unit of execution in CumulusCI.

Subclass BaseTask or a descendant to define custom task logic
"""
import contextlib
import logging
import os
import re
import threading
import time
from contextlib import nullcontext
from typing import Any, Callable, Dict, List, Optional, Type, Union

from pydantic.error_wrappers import ValidationError

from cumulusci import __version__
from cumulusci.core.config import TaskConfig
from cumulusci.core.config.org_config import OrgConfig
from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.debug import DebugMode, get_debug_mode
from cumulusci.core.exceptions import (
    ServiceNotConfigured,
    ServiceNotValid,
    TaskOptionsError,
    TaskRequiresSalesforceOrg,
)
from cumulusci.core.flowrunner import FlowCoordinator, StepSpec, StepVersion
from cumulusci.utils import cd
from cumulusci.utils.logging import redirect_output_to_logger
from cumulusci.utils.metaprogramming import classproperty
from cumulusci.utils.options import CCIOptions, ReadOnlyOptions

CURRENT_TASK = threading.local()

PROJECT_CONFIG_RE = re.compile(r"\$project_config.(\w+)")
CAPTURE_TASK_OUTPUT = os.environ.get("CAPTURE_TASK_OUTPUT")


@contextlib.contextmanager
def stacked_task(task):
    if not hasattr(CURRENT_TASK, "stack"):
        CURRENT_TASK.stack = []
    CURRENT_TASK.stack.append(task)
    try:
        yield
    finally:
        CURRENT_TASK.stack.pop()


class BaseTask:
    """BaseTask provides the core execution logic for a Task

    Subclass BaseTask and provide a `_run_task()` method with your
    code.
    """

    task_docs: str = ""
    Options: Optional[Type[CCIOptions]] = None
    salesforce_task: bool = False  # Does this task require a salesforce org?
    name: Optional[str]
    stepnum: Optional[StepVersion]
    result: Any
    return_values: dict
    debug_mode: DebugMode
    logger: logging.Logger
    options: dict

    poll_complete: bool
    poll_count: int
    poll_interval_level: int
    poll_interval_s: int

    def __init__(
        self,
        project_config: BaseProjectConfig,
        task_config: TaskConfig,
        org_config: Optional[OrgConfig] = None,
        flow: Optional[FlowCoordinator] = None,
        name: Optional[str] = None,
        stepnum: Optional[StepVersion] = None,
        logger: Optional[logging.Logger] = None,
        **kwargs,
    ):
        self.project_config = project_config
        self.task_config = task_config
        self.org_config = org_config

        self._reset_poll()

        # dict of return_values that can be used by task callers
        self.return_values = {}
        # simple result object for introspection, often a return_code
        self.result = None
        # the flow for this task execution
        self.flow = flow
        # the task's name in the flow
        self.name = name
        # the task's stepnumber in the flow
        self.stepnum = stepnum

        self.debug_mode = get_debug_mode()
        if logger:
            self.logger = logger
        else:
            self._init_logger()

        self._init_options(kwargs)
        self._validate_options()

    def _init_logger(self):
        """Initializes self.logger"""
        if self.flow:
            self.logger = self.flow.logger.getChild(self.__class__.__name__)
        else:
            self.logger = logging.getLogger(__name__)

    @classproperty
    def task_options(cls):  # type: ignore  -- doesn't like the signature override below
        "Convert Options dict into old fashioned task_options syntax"
        if cls.Options:
            return cls.Options.as_task_options()
        else:
            return {}

    task_options: Union[classproperty, Dict]

    def _init_options(self, kwargs):
        """Initializes self.options"""
        if self.task_config.options is None:
            self.options = {}
        else:
            self.options = self.task_config.options.copy()

        if kwargs:
            self.options.update(kwargs)

        # Handle dynamic lookup of project_config values via $project_config.attr
        def process_options(option):
            if isinstance(option, str):
                return PROJECT_CONFIG_RE.sub(
                    lambda match: str(self.project_config.lookup(match.group(1), None)),
                    option,
                )
            elif isinstance(option, dict):
                processed_dict = {}
                for key, value in option.items():
                    processed_dict[key] = process_options(value)
                return processed_dict
            elif isinstance(option, list):
                processed_list = []
                for item in option:
                    processed_list.append(process_options(item))
                return processed_list
            else:
                return option

        self.options = process_options(self.options)

        if self.Options:
            try:
                specials = ["debug_before", "debug_after", "no_prompt"]
                options_without_specials = {
                    opt: val for opt, val in self.options.items() if opt not in specials
                }
                self.parsed_options = self.Options(**options_without_specials)
                self.options = ReadOnlyOptions(self.options)
            except ValidationError as e:
                try:
                    errors = [
                        f"Error in '{error['loc'][0]}' option: '{error['msg']}'"
                        for error in e.errors()
                    ]
                    plural = "s" if len(errors) > 1 else ""
                    errorstrs = ", ".join(errors)
                    message = f"Task Options Error{plural}: {errorstrs}"
                except (AttributeError, IndexError):
                    message = f"Task Options Error: {e.errors()}"
                if "extra fields not permitted" in message:
                    message = message.replace("extra fields", "extra options")
                raise TaskOptionsError(message) from e
            except (TaskOptionsError, TypeError) as e:
                raise TaskOptionsError(
                    f"Task Options Error: Error in '{self.options}' option: '{e}'"
                )

    def _validate_options(self):
        missing_required = []
        for name, config in list(self.task_options.items()):
            if config.get("required") is True and name not in self.options:
                missing_required.append(name)

        if missing_required:
            required_opts = ",".join(missing_required)
            raise TaskOptionsError(
                f"{self.__class__.__name__} requires the options ({required_opts}) and no values were provided"
            )

    def _update_credentials(self):
        """Override to do any logic to refresh credentials"""
        pass

    def _init_task(self):
        """Override to implement dynamic logic for initializing the task."""
        pass

    def __call__(self) -> dict:
        if self.salesforce_task and not self.org_config:
            raise TaskRequiresSalesforceOrg(
                "This task requires a salesforce org. "
                "Use org default <name> to set a default org "
                "or pass the org name with the --org option"
            )
        self._update_credentials()
        self._init_task()

        with stacked_task(self):
            self.working_path = os.getcwd()
            path = self.project_config.repo_root if self.project_config else None
            with cd(path):
                with (
                    redirect_output_to_logger(self.logger)
                    if CAPTURE_TASK_OUTPUT
                    else nullcontext()
                ):
                    self._log_begin()
                    self.result = self._run_task()
                    return self.return_values

    def _run_task(self) -> Any:
        """Subclasses should override to provide their implementation"""
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _log_begin(self):
        """Log the beginning of the task execution"""
        self.logger.info(f"Beginning task: {self.__class__.__name__}")
        if self.salesforce_task and not self.flow:
            self.logger.info(f"As user: {self.org_config.username}")
            self.logger.info(f"In org: {self.org_config.org_id}")
        self.logger.info("")

    def _retry(self):
        while True:
            try:
                self._try()
                break
            except Exception as e:
                if not (self.options["retries"] and self._is_retry_valid(e)):
                    raise
                if self.options["retry_interval"]:
                    self.logger.warning(
                        f"Sleeping for {self.options['retry_interval']} seconds before retry..."
                    )
                    time.sleep(self.options["retry_interval"])
                    if self.options["retry_interval_add"]:
                        self.options["retry_interval"] += self.options[
                            "retry_interval_add"
                        ]
                self.options["retries"] -= 1
                self.logger.warning(
                    f"Retrying ({self.options['retries']} attempts remaining)"
                )

    def _try(self):
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _is_retry_valid(self, e: Exception) -> bool:
        return True

    def _reset_poll(self):
        self.poll_complete = False
        self.poll_count = 0
        self.poll_interval_level = 0
        self.poll_interval_s = 1

    def _poll(self):
        """poll for a result in a loop"""
        while True:
            self.poll_count += 1
            self._poll_action()
            if self.poll_complete:
                break
            time.sleep(self.poll_interval_s)
            self._poll_update_interval()

    def _poll_action(self):
        """
        Poll something and process the response.
        Set `self.poll_complete = True` to break polling loop.
        """
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _poll_update_interval(self):
        """update the polling interval to be used next iteration"""
        # Increase by 1 second every 3 polls
        if self.poll_count // 3 > self.poll_interval_level:
            self.poll_interval_level += 1
            self.poll_interval_s += 1
            self.logger.info(
                "Increased polling interval to %d seconds", self.poll_interval_s
            )

    def freeze(self, step: StepSpec) -> List[dict]:
        ui_step = {
            "name": self.task_config.name or self.name,
            "kind": "other",
            "is_required": True,
        }
        ui_step.update(self.task_config.config.get("ui_options", {}))
        task_config = {
            "options": self.options,
            "checks": self.task_config.checks or [],
        }
        ui_step.update(
            {
                "path": step.path,
                "step_num": str(step.step_num),
                "task_class": self.task_config.class_path,
                "task_config": task_config,
                "source": step.project_config.source.frozenspec,
            }
        )
        return [ui_step]


class BaseSalesforceTask(BaseTask):
    """Base for tasks that need a Salesforce org"""

    name = "BaseSalesforceTask"
    salesforce_task = True

    def _get_client_name(self):
        try:
            app = self.project_config.keychain.get_service("connectedapp")
            return app.client_id
        except (ServiceNotValid, ServiceNotConfigured):
            return f"CumulusCI/{__version__}"

    def _run_task(self):
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _update_credentials(self):
        with self.org_config.save_if_changed():
            self.org_config.refresh_oauth_token(self.project_config.keychain)

    def _validate_and_inject_namespace_prefixes(
        self,
        should_inject_namespaces: bool,
        sobjects_to_validate: List[str],
        operation_to_validate: str,
    ) -> List[str]:
        """Perform namespace injection and ensure that we can successfully access all of the selected objects."""

        global_describe = {
            entry["name"]: entry
            for entry in self.org_config.salesforce_client.describe()["sobjects"]
        }

        # Namespace injection
        if should_inject_namespaces and self.project_config.project__package__namespace:

            def inject(element: str):
                return f"{self.project_config.project__package__namespace}__{element}"

            sobjects = [
                self.fixup_sobject_name(inject, sobject, global_describe)
                for sobject in sobjects_to_validate
            ]
        else:
            sobjects = sobjects_to_validate

        # Validate CRUD
        non_accessible_objects = [
            s
            for s in sobjects
            if not (s in global_describe and global_describe[s][operation_to_validate])
        ]
        if non_accessible_objects:
            raise TaskOptionsError(
                f"The objects {', '.join(non_accessible_objects)} are not present or not {operation_to_validate}."
            )
        return sobjects

    def fixup_sobject_name(
        self,
        injection_func: Callable[[str], str],
        sobject: str,
        global_describe: Dict[str, dict],
    ) -> str:
        def _is_injectable(element: str) -> bool:
            return element.count("__") == 1

        if not _is_injectable(sobject):
            return sobject

        injected = injection_func(sobject)
        if sobject in global_describe and injected in global_describe:
            self.logger.warning(
                f"Both {sobject} and {injected} are present in the target org. Using {sobject}."
            )

        if sobject not in global_describe and injected in global_describe:
            return injected
        else:
            return sobject



import requests

from cumulusci.core.utils import process_bool_arg
from cumulusci.salesforce_api import mc_soap_envelopes as envelopes

from .base import BaseMarketingCloudTask


class CreateSubscriberAttribute(BaseMarketingCloudTask):
    task_options = {  # TODO: should use `class Options instead`
        "attribute_name": {
            "description": "The name of the Subscriber Attribute to deploy via the Marketing Cloud API.",
            "required": True,
        },
    }

    def _run_task(self):
        attribute_name = self.options["attribute_name"]
        # get soap envelope
        envelope = envelopes.CREATE_SUBSCRIBER_ATTRIBUTE
        # fill the merge fields
        envelope = envelope.format(
            soap_instance_url=self.mc_config.soap_instance_url,
            access_token=self.mc_config.access_token,
            attribute_name=attribute_name,
        )
        response = requests.post(
            f"{self.mc_config.soap_instance_url}Service.asmx",
            data=envelope.encode("utf-8"),
            headers={"Content-Type": "text/xml; charset=utf-8"},
        )
        self._check_soap_response(response)
        self.logger.info(
            f"Successfully created subscriber attribute: {attribute_name}."
        )
        self.return_values = {"success": True}


class CreateUser(BaseMarketingCloudTask):
    task_options = {  # TODO: should use `class Options instead`
        "parent_bu_mid": {
            "description": "Specify the MID for Parent BU.",
            "required": True,
        },
        "default_bu_mid": {
            "description": "Set MID for BU to use as default (can be same as the parent).",
            "required": True,
        },
        "external_key": {
            "description": "Set the User's external key.",
            "required": False,
        },
        "user_name": {
            "description": "Set the User's name. Not the same as their username.",
            "required": False,
        },
        "user_email": {
            "description": "Set the User's email.",
            "required": True,
        },
        "user_password": {
            "description": "Set the User's password.",
            "required": True,
            "sensitive": True,
        },
        "user_username": {
            "description": "Set the User's username. Not the same as their name.",
            "required": True,
        },
        "role_id": {
            "description": "Assign a Role to the new User, specified as an ID. IDs for system defined roles located here: https://developer.salesforce.com/docs/atlas.en-us.noversion.mc-apis.meta/mc-apis/setting_user_permissions_via_the_web_services_api.htm",
            "required": False,
        },
        "activate_if_existing": {
            "description": "Activate the user if it already exists in an inactive state. Default: False",
            "required": False,
        },
    }

    def _run_task(self):
        # get soap envelope
        envelope = envelopes.CREATE_USER
        # fill the merge fields
        # check for optional parameters, construct xml nodes as required
        external_key = self.options.get("external_key", "")
        if external_key != "":
            external_key = f"<CustomerKey>{external_key}</CustomerKey>"

        user_username = self.options.get("user_username")

        user_name = self.options.get("user_name", "")
        if user_name != "":
            user_name = f"<Name>{user_name}</Name>"

        role_id = self.options.get("role_id", "")
        if role_id != "":
            role_id = f"<UserPermissions><ID>{role_id}</ID></UserPermissions>"

        active_flag = (
            "<ActiveFlag>true</ActiveFlag>"
            if process_bool_arg(self.options.get("activate_if_existing"))
            else ""
        )

        envelope = envelope.format(
            soap_instance_url=self.mc_config.soap_instance_url,
            access_token=self.mc_config.access_token,
            parent_bu_mid=self.options.get("parent_bu_mid"),
            default_bu_mid=self.options.get("default_bu_mid"),
            external_key=external_key,
            user_name=user_name,
            user_email=self.options.get("user_email"),
            user_password=self.options.get("user_password"),
            user_username=user_username,
            role_id=role_id,
            active_flag=active_flag,
        )
        response = requests.post(
            f"{self.mc_config.soap_instance_url}Service.asmx",
            data=envelope.encode("utf-8"),
            headers={"Content-Type": "text/xml; charset=utf-8"},
        )
        self._check_soap_response(response)
        self.logger.info(f"Successfully created User: {user_username}.")
        self.return_values = {"success": True}


class UpdateUserRole(BaseMarketingCloudTask):
    task_options = {  # TODO: should use `class Options instead`
        "account_mid": {
            "description": "Specify the Account MID.",
            "required": True,
        },
        "external_key": {
            "description": "Specify the User's external key.",
            "required": False,
        },
        "user_name": {
            "description": "Specify the User's name. Not the same as their username.",
            "required": False,
        },
        "user_email": {
            "description": "Specify the User's email.",
            "required": True,
        },
        "user_password": {
            "description": "Specify the User's password.",
            "required": True,
            "sensitive": True,
        },
        "role_id": {
            "description": "Assign a Role to the User, specified as an ID. IDs for system defined roles located here: https://developer.salesforce.com/docs/atlas.en-us.noversion.mc-apis.meta/mc-apis/setting_user_permissions_via_the_web_services_api.htm",
            "required": True,
        },
    }

    def _run_task(self):
        # get soap envelope
        envelope = envelopes.UPDATE_USER_ROLE
        # fill the merge fields
        # check for optional parameters, construct xml nodes as required
        external_key = self.options.get("external_key", "")
        if external_key != "":
            external_key = f"<CustomerKey>{external_key}</CustomerKey>"

        user_name = self.options.get("user_name", "")
        if user_name != "":
            user_name = f"<Name>{user_name}</Name>"

        envelope = envelope.format(
            soap_instance_url=self.mc_config.soap_instance_url,
            access_token=self.mc_config.access_token,
            account_mid=self.options.get("account_mid"),
            external_key=external_key,
            user_name=user_name,
            user_email=self.options.get("user_email"),
            user_password=self.options.get("user_password"),
            role_id=self.options.get("role_id"),
        )
        response = requests.post(
            f"{self.mc_config.soap_instance_url}Service.asmx",
            data=envelope.encode("utf-8"),
            headers={"Content-Type": "text/xml; charset=utf-8"},
        )
        self._check_soap_response(response)
        user_name = self.options.get("user_name")
        self.logger.info(f"Successfully updated role for User: {user_name}.")
        self.return_values = {"success": True}



import functools
import re

from robot.api import logger
from robot.libraries.BuiltIn import BuiltIn
from SeleniumLibrary.errors import ElementNotFound

from cumulusci.core.utils import dictmerge

"""
This module supports managing multiple location strategies. It
works like this:

0. Locators are stored in a global LOCATORS dictionary. The keys
   are the locator prefixes, the values are dictionaries containing
   the locators

1. Libraries can register a dictionary of locators with a prefix
   (eg: register_locators("npsp", {...})). These get added to
   LOCATORS

2. Open Test Browser calls selenium's add_location_strategy for each
   registered set of locators. (Note: Location strategies cannot be added before a
   browser is open).

3. Keywords can use dot notation to refer to locators. A colon
   separates the prefix from the locator, and a locator from a
   comma-separated string of arguments

Example:

    from a keyword library:

    | from cumulusci.robotframework.locator_manager import register_locators
    |
    | register_locators("example", {"dialog": {"button": "//button[text()='{}']"}})

    in a test:

    | page should contain element  example:dialog.button:Save

    ... will result in "//button[text()='Save']" being passed
    to 'page should contain element'

"""

LOCATORS = {}


def register_locators(prefix, locators):
    """Register locators to be used with a custom locator strategy

    If the prefix is already known, the locators will be merged with
    the dictionary we already have.

    """
    if prefix in LOCATORS:
        logger.debug(f"merging keywords for prefix {prefix}")
        dictmerge(LOCATORS[prefix], locators)
    else:
        logger.debug(f"registering keywords for prefix {prefix}")
        LOCATORS[prefix] = locators


def add_location_strategies():
    """Call selenium's add_location_strategy keyword for all strategies"""

    # selenium throws an error if the location strategy already
    # exists, so we use a flag to make sure this code is called
    # only once.
    selenium = BuiltIn().get_library_instance("SeleniumLibrary")
    for (prefix, strategy) in LOCATORS.items():
        try:
            logger.debug(f"adding location strategy for '{prefix}'")
            selenium.add_location_strategy(
                prefix, functools.partial(locate_element, prefix)
            )
        except Exception as e:
            logger.debug(f"unable to register locators: {e}")


def locate_element(prefix, parent, locator, tag, constraints):
    """Translate a custom locator specification into an actual locator

    Our custom locators are of the form "p:x.y.z:a,b" where:

    - p is a short prefix (eg: sf, eda, sal),
    - x,y,z are keys to a locator dictionary (eg: locators['x']['y']['z'])
    - a,b are positional parameters passed to the string.format method
      when converting the custom locator into an actual locator

    A locator string (eg: locators['x']['y']['z']) can have substitution
    fields in it (eg: "a[@title='{}']"). These fields will be replaced
    with the positional parameters. It is possible for these fields to be
    named (eg: "a[@title='{title}'"), though we don't support named
    parameters.

    If the substitution fields are named, each unique name will be
    associated with a positional argument, in order. For example, if
    the arguments are "a,b" and if the locator string is something
    like "//{foo}|//{foo}/{bar}", then the first argument (a) will be
    assigned to the first namef field (foo) and the second argument (b)
    will be assigned to the second named field (bar).

    """

    selenium = BuiltIn().get_library_instance("SeleniumLibrary")
    loc = translate_locator(prefix, locator)
    logger.info(f"locator: '{prefix}:{locator}' => '{loc}'")

    try:
        elements = selenium.get_webelements(loc)
    except Exception as e:
        # The SeleniumLibrary documentation doesn't say, but I'm
        # pretty sure we should return an empty list rather than
        # throwing an error in this case. If we throw an error, that
        # prevents the custom locators from being used negatively (eg:
        # Page should not contain element custom:whatever).
        logger.debug(f"caught exception in locate_element: {e}")
        elements = []
    return elements


def translate_locator(prefix, locator):
    """Return the translated locator

    This uses the passed-in prefix and locator to find the
    proper element in the LOCATORS dictionary, and then formats it
    with any arguments that were part of the locator.

    See the docstring for `locate_element` for a description of how
    positional arguments are applied to named format fields.

    """

    if ":" in locator:
        (path, argstring) = locator.split(":", 1)
    else:
        path = locator
        argstring = ""

    loc = LOCATORS[prefix]
    breadcrumbs = []
    try:
        for key in path.split("."):
            breadcrumbs.append(key)
            # this assumes that loc is a dictionary rather than a
            # string. If we've hit the leaf node of the locator and
            # there are still more keys, this will fail with a TypeError
            loc = loc[key.strip()]

    except (KeyError, TypeError):
        # TypeError: if the user passes in foo.bar.baz, but foo or foo.bar
        # resolves to a string rather than a nested dict.
        # KeyError if user passes in foo.bar and either 'foo' or 'bar' isn't
        # a valid key for a nested dictionary
        breadcrumb_path = ".".join(breadcrumbs)
        raise ElementNotFound(f"locator {prefix}:{breadcrumb_path} not found")

    if not isinstance(loc, str):
        raise TypeError(f"Expected locator to be of type string, but was {type(loc)}")

    try:
        # args is still a string, so split on ","
        # This means that arguments can't have commas in them, but I'm not sure
        # that will be a problem. If we find a case where it's a problem we can
        # do more sophisticated parsing.
        args = [arg.strip() for arg in argstring.split(",")] if argstring else []
        loc = apply_formatting(loc, args)
    except IndexError:
        raise Exception("Not enough arguments were supplied")
    return loc


def apply_formatting(locator, args):
    """Apply formatting to the locator

    If there are no named fields in the locator this is just a simple
    call to .format. However, some locators have named fields, and we
    don't support named arguments to keep the syntax simple, so we
    need to map positional arguments to named arguments before calling
    .format.

    Example:

    Given the locator "//*[a[@title='{title}'] or
    button[@name='{title}']]//{tag}" and args of ['foo', 'bar'], we'll
    pop 'foo' and 'bar' off of the argument list and assign them to
    the kwargs keys 'title' and 'tag'.

    """
    kwargs = {}
    for match in re.finditer(r"\{([^}]+)\}", locator):
        name = match.group(1)
        if name and name not in kwargs:
            kwargs[name] = args.pop(0)
    return locator.format(*args, **kwargs)



from abc import ABC, abstractclassmethod
from typing import Dict

from cumulusci.core.config import ServiceConfig


class OAuth2ServiceConfig(ServiceConfig, ABC):
    """Base class for services that require an OAuth2 Client
    for establishing a connection."""

    @abstractclassmethod  # type: ignore
    def connect(cls) -> Dict:  # type: ignore
        """This method is called when the service is first connected
        via `cci service connect`. This method should perform the necessary
        OAuth flow and return a dict of values that the service would like
        stored in the services `config` dict."""
        raise NotImplementedError



from cumulusci.tasks.base_source_control_task import BaseSourceControlTask


class CreatePackageDataFromCommitStatus(BaseSourceControlTask):
    task_options = {
        "state": {
            "description": "sha of the commit",
            "required": True,
        },
        "context": {
            "description": "Name of the commit status context",
            "required": True,
        },
        "commit_id": {
            "description": "sha of the commit",
        },
        "description": {
            "description": "Description of the commit status",
        },
        "target_url": {
            "description": "URL to associate with the commit status",
        },
    }

    def _run_task(self):
        repo = self.get_repo()
        commit_sha = self.options["commit_id"] or self.project_config.repo_commit or ""

        commit = repo.create_commit_status(
            commit_sha,
            state=self.options["state"],
            context=self.options["context"],
            description=self.options.get("description"),
            target_url=self.options.get("target_url"),
        )

        self.return_values = {"commit_id": commit.sha}



from unittest.mock import Mock

from cumulusci.tasks.preflight.licenses import (
    GetAssignableLicenses,
    GetAssignablePermissionSets,
    GetAvailableLicenses,
    GetAvailablePermissionSetLicenses,
    GetAvailablePermissionSets,
    GetPermissionLicenseSetAssignments,
)
from cumulusci.tasks.salesforce.tests.util import create_task


class TestLicensePreflights:
    def test_license_preflight(self):
        task = create_task(GetAvailableLicenses, {})
        task._init_api = Mock()
        task._init_api.return_value.query.return_value = {
            "totalSize": 2,
            "records": [
                {
                    "Id": "L1",
                    "LicenseDefinitionKey": "TEST1",
                    "TotalLicenses": 100,
                    "UsedLicenses": 90,
                },
                {
                    "Id": "L2",
                    "LicenseDefinitionKey": "TEST2",
                    "TotalLicenses": 100,
                    "UsedLicenses": 100,
                },
            ],
        }

        task()
        task._init_api.return_value.query.assert_called_once_with(
            "SELECT Id, LicenseDefinitionKey, TotalLicenses, UsedLicenses FROM UserLicense WHERE Status = 'Active'"
        )

        assert task.return_values == ["TEST1", "TEST2"]

    def test_assignable_license_preflight(self):
        task = create_task(GetAssignableLicenses, {})
        task._init_api = Mock()
        task._init_api.return_value.query.return_value = {
            "totalSize": 2,
            "records": [
                {
                    "Id": "L1",
                    "LicenseDefinitionKey": "TEST1",
                    "TotalLicenses": 100,
                    "UsedLicenses": 90,
                },
                {
                    "Id": "L2",
                    "LicenseDefinitionKey": "TEST2",
                    "TotalLicenses": 100,
                    "UsedLicenses": 100,
                },
            ],
        }

        task()
        task._init_api.return_value.query.assert_called_once_with(
            "SELECT Id, LicenseDefinitionKey, TotalLicenses, UsedLicenses FROM UserLicense WHERE Status = 'Active'"
        )
        # Only TEST1 assignable licenses
        assert task.return_values == ["TEST1"]

    def test_psl_preflight(self):
        task = create_task(GetAvailablePermissionSetLicenses, {})
        task._init_api = Mock()

        task._init_api.return_value.query.return_value = {
            "totalSize": 2,
            "records": [
                {"PermissionSetLicenseKey": "TEST1"},
                {"PermissionSetLicenseKey": "TEST2"},
            ],
        }
        task()

        task._init_api.return_value.query.assert_called_once_with(
            "SELECT PermissionSetLicenseKey FROM PermissionSetLicense WHERE Status = 'Active'"
        )
        assert task.return_values == ["TEST1", "TEST2"]

    def test_assigned_permsetlicense_preflight(self):
        task = create_task(GetPermissionLicenseSetAssignments, {})
        task._init_api = Mock()
        task._init_api.return_value.query_all.return_value = {
            "totalSize": 2,
            "done": True,
            "records": [
                {
                    "PermissionSetLicense": {
                        "MasterLabel": "Document Checklist",
                        "DeveloperName": "DocumentChecklist",
                    },
                },
                {
                    "PermissionSetLicense": {
                        "MasterLabel": "Einstein Analytics Plus Admin",
                        "DeveloperName": "EinsteinAnalyticsPlusAdmin",
                    },
                },
            ],
        }
        task()

        task._init_api.return_value.query_all.assert_called_once_with(
            "SELECT PermissionSetLicense.DeveloperName FROM PermissionSetLicenseAssign WHERE AssigneeId = 'USER_ID'"
        )
        assert task.return_values == [
            "DocumentChecklist",
            "EinsteinAnalyticsPlusAdmin",
        ]

    def test_permsets_preflight(self):
        task = create_task(GetAvailablePermissionSets, {})
        task._init_api = Mock()

        task._init_api.return_value.query_all.return_value = {
            "totalSize": 2,
            "records": [
                {"Name": "TEST1"},
                {"Name": "TEST2"},
            ],
        }
        task()

        task._init_api.return_value.query_all.assert_called_once_with(
            "SELECT Name FROM PermissionSet"
        )
        assert task.return_values == ["TEST1", "TEST2"]

    def test_assignable_permsets_preflight(self):
        task = create_task(GetAssignablePermissionSets, {})
        task._init_api = Mock()
        task._init_api.return_value.query.return_value = {
            "totalSize": 2,
            "records": [
                {
                    "Id": "L1",
                    "LicenseDefinitionKey": "TEST1",
                    "TotalLicenses": 100,
                    "UsedLicenses": 90,
                },
                {
                    "Id": "L2",
                    "LicenseDefinitionKey": "TEST2",
                    "TotalLicenses": 100,
                    "UsedLicenses": 100,
                },
            ],
        }
        task._init_api.return_value.query_all.return_value = {
            "totalSize": 3,
            "records": [
                {"LicenseId": "L1", "Name": "TEST1"},
                {"LicenseId": "L2", "Name": "TEST2"},
                {"LicenseId": None, "Name": "TEST3"},
            ],
        }
        task()

        task._init_api.return_value.query.assert_called_once_with(
            "SELECT Id, LicenseDefinitionKey, TotalLicenses, UsedLicenses FROM UserLicense WHERE Status = 'Active'"
        )
        task._init_api.return_value.query_all.assert_called_once_with(
            "SELECT LicenseId, Name FROM PermissionSet"
        )
        assert task.return_values == ["TEST1", "TEST3"]



import json
from collections import defaultdict
from datetime import datetime
from pathlib import Path

import click

from cumulusci.core.exceptions import FlowNotFoundError
from cumulusci.core.utils import format_duration
from cumulusci.utils import document_flow, flow_ref_title_and_intro
from cumulusci.utils.yaml.safer_loader import load_yaml_data

from .runtime import pass_runtime
from .ui import CliTable
from .utils import group_items


@click.group("flow", help="Commands for finding and running flows for a project")
def flow():
    pass


@flow.command(name="doc", help="Exports RST format documentation for all flows")
@click.option(
    "--project", "project", is_flag=True, help="Include project-specific flows only"
)
@pass_runtime(require_project=False, require_keychain=True)
def flow_doc(runtime, project=False):
    flow_info_path = Path(__file__, "..", "..", "..", "docs", "flows.yml").resolve()
    with open(flow_info_path, "r", encoding="utf-8") as f:
        flow_info = load_yaml_data(f)
    click.echo(flow_ref_title_and_intro(flow_info["intro_blurb"]))
    flow_info_groups = list(flow_info["groups"].keys())

    universal_flows = runtime.universal_config.list_flows()
    if project:
        flows = [
            flow
            for flow in runtime.project_config.list_flows()
            if flow not in universal_flows
        ]
    else:
        flows = universal_flows
    flows_by_group = group_items(flows)
    flow_groups = sorted(
        flows_by_group.keys(),
        key=lambda group: flow_info_groups.index(group)
        if group in flow_info_groups
        else 100,
    )

    for group in flow_groups:
        click.echo(f"{group}\n{'-' * len(group)}")
        if group in flow_info["groups"]:
            click.echo(flow_info["groups"][group]["description"])

        for flow in sorted(flows_by_group[group]):
            flow_name, flow_description = flow
            try:
                flow_coordinator = runtime.get_flow(flow_name)
            except FlowNotFoundError as e:
                raise click.UsageError(str(e))

            additional_info = None
            if flow_name in flow_info.get("flows", {}):
                additional_info = flow_info["flows"][flow_name]["rst_text"]

            click.echo(
                document_flow(
                    flow_name,
                    flow_description,
                    flow_coordinator,
                    additional_info=additional_info,
                )
            )
            click.echo("")


@flow.command(name="list", help="List available flows for the current context")
@click.option("--plain", is_flag=True, help="Print the table using plain ascii.")
@click.option("--json", "print_json", is_flag=True, help="Print a json string")
@pass_runtime(require_project=False)
def flow_list(runtime, plain, print_json):
    plain = plain or runtime.universal_config.cli__plain_output
    flows = runtime.get_available_flows()
    if print_json:
        click.echo(json.dumps(flows))
        return None

    flow_groups = group_items(flows)
    for group, flows in flow_groups.items():
        data = [["Flow", "Description"]]
        data.extend(sorted(flows))
        table = CliTable(
            data,
            group,
        )
        table.echo(plain)

    click.echo(
        "Use "
        + click.style("cci flow info <flow_name>", bold=True)
        + " to get more information about a flow."
    )


@flow.command(name="info", help="Displays information for a flow")
@click.argument("flow_name")
@pass_runtime(require_keychain=True)
def flow_info(runtime, flow_name):
    try:
        coordinator = runtime.get_flow(flow_name)
        output = coordinator.get_summary(verbose=True)
        click.echo(output)
    except FlowNotFoundError as e:
        raise click.UsageError(str(e))


@flow.command(name="run", help="Runs a flow")
@click.argument("flow_name")
@click.option(
    "--org",
    help="Specify the target org.  By default, runs against the current default org",
)
@click.option(
    "--delete-org",
    is_flag=True,
    help="If set, deletes the scratch org after the flow completes",
)
@click.option(
    "--debug", is_flag=True, help="Drops into pdb, the Python debugger, on an exception"
)
@click.option(
    "-o",
    nargs=2,
    multiple=True,
    help="Pass task specific options for the task as '-o taskname__option value'.  You can specify more than one option by using -o more than once.",
)
@click.option(
    "--no-prompt",
    is_flag=True,
    help="Disables all prompts.  Set for non-interactive mode use such as calling from scripts or CI systems",
)
@pass_runtime(require_keychain=True)
def flow_run(runtime, flow_name, org, delete_org, debug, o, no_prompt):

    # Get necessary configs
    org, org_config = runtime.get_org(org)
    if delete_org and not org_config.scratch:
        raise click.UsageError("--delete-org can only be used with a scratch org")

    # Parse command line options
    options = defaultdict(dict)
    if o:
        for key, value in o:
            if "__" in key:
                task_name, option_name = key.split("__")
                options[task_name][option_name] = value
            else:
                raise click.UsageError(
                    "-o option for flows should contain __ to split task name from option name."
                )

    # Create the flow and handle initialization exceptions
    try:
        coordinator = runtime.get_flow(flow_name, options=options)
        start_time = datetime.now()
        coordinator.run(org_config)
        duration = datetime.now() - start_time
        click.echo(f"Ran {flow_name} in {format_duration(duration)}")
    except Exception:
        runtime.alert(f"Flow error: {flow_name}")
        raise
    finally:
        # Delete the scratch org if --delete-org was set
        if delete_org:
            try:
                org_config.delete_org()
            except Exception as e:
                click.echo(
                    "Scratch org deletion failed.  Ignoring the error below to complete the flow:"
                )
                click.echo(str(e))

    runtime.alert(f"Flow Complete: {flow_name}")



import contextlib
import io
import json
import logging
import os
import pathlib
import platform
import sys
import typing as T
from os import PathLike
from zipfile import ZipFile

import sarge

# Fix for TextIOWrapper flush issue with sarge.Capture objects
# Add flush method to sarge.Capture to prevent AttributeError during garbage collection
if not hasattr(sarge.Capture, 'flush'):
    def _capture_flush(self):
        """No-op flush method for sarge.Capture compatibility with TextIOWrapper"""
        pass
    sarge.Capture.flush = _capture_flush

from cumulusci.core.enums import StrEnum
from cumulusci.core.exceptions import SfdxOrgException
from cumulusci.utils import temporary_dir

logger = logging.getLogger(__name__)


def sfdx(
    command,
    username=None,
    log_note=None,
    access_token=None,
    args: T.Optional[T.List[str]] = None,
    env=None,
    capture_output=True,
    check_return=False,
):
    """Call an sfdx command and capture its output.

    Be sure to quote user input that is part of the command using `shell_quote`.

    Returns a `sarge` Command instance with returncode, stdout, stderr
    """
    command = f"sf {command}"
    if args is not None:
        for arg in args:
            command += " " + shell_quote(arg)
    if username:
        command += f" -o {shell_quote(username)}"
    if log_note:
        logger.info(f"{log_note} with command: {command}")
    # Avoid logging access token
    if access_token:
        command += f" -o {shell_quote(access_token)}"
    env = env or {}
    p = sarge.Command(
        command,
        stdout=sarge.Capture(buffer_size=-1) if capture_output else None,
        stderr=sarge.Capture(buffer_size=-1) if capture_output else None,
        shell=True,
        env={**env, "SFDX_TOOL": "CCI"},
    )
    p.run()
    if capture_output:
        p.stdout_text = io.TextIOWrapper(p.stdout, encoding=sys.stdout.encoding)
        p.stderr_text = io.TextIOWrapper(p.stderr, encoding=sys.stdout.encoding)
    if check_return and p.returncode:
        message = f"Command exited with return code {p.returncode}"
        if capture_output:
            message += f":\n{p.stderr_text.read()}"
        raise Exception(message)
    return p


def shell_quote(s: str):
    if platform.system() == "Windows":
        assert isinstance(s, str)
        if not s:
            result = '""'
        elif '"' not in s:
            result = s
            if " " in result:
                result = f'"{result}"'
        else:
            escaped = s.replace('"', r"\"")
            result = f'"{escaped}"'

        return result
    else:
        return sarge.shell_quote(s)


def get_default_devhub_username():
    p = sfdx(
        "config get target-dev-hub --json",
        log_note="Getting default Dev Hub username from sfdx",
        check_return=True,
    )
    result = json.load(p.stdout_text)
    if "result" not in result or "value" not in result["result"][0]:
        raise SfdxOrgException(
            "No sf config found for target-dev-hub. "
            "Please use the sf config set to set the target-dev-hub and run again."
        )
    username = result["result"][0]["value"]
    return username


class SourceFormat(StrEnum):
    SFDX = "SFDX"
    MDAPI = "MDAPI"


def get_source_format_for_path(path: T.Optional[PathLike]) -> SourceFormat:
    if pathlib.Path(path or pathlib.Path.cwd(), "package.xml").exists():
        return SourceFormat.MDAPI

    return SourceFormat.SFDX


def get_source_format_for_zipfile(
    zf: ZipFile, subfolder: T.Optional[str]
) -> SourceFormat:
    namelist = zf.namelist()

    target_name = str(pathlib.PurePosixPath(subfolder or "", "package.xml"))

    if target_name in namelist:
        return SourceFormat.MDAPI

    return SourceFormat.SFDX


@contextlib.contextmanager
def convert_sfdx_source(
    path: T.Optional[PathLike], name: T.Optional[str], logger: logging.Logger
):
    mdapi_path = None
    with contextlib.ExitStack() as stack:
        # Convert SFDX -> MDAPI format if path exists but does not have package.xml
        if (
            len(os.listdir(path))  # path is None -> CWD
            and get_source_format_for_path(path) is SourceFormat.SFDX
        ):
            logger.info("Converting from SFDX to MDAPI format.")
            mdapi_path = stack.enter_context(temporary_dir(chdir=False))
            args = ["-d", mdapi_path]
            if path:
                # No path means convert default package directory in the CWD
                args += ["-r", str(path)]
            if name:
                args += ["-n", name]
            sfdx(
                "project convert source",
                args=args,
                capture_output=True,
                check_return=True,
            )

        yield mdapi_path or path
